#!/usr/bin/env Rscript

library(RMySQL)
library(plyr)
options(width=400)

readData <- function(tableName) {
	cat("Reading data from local database…\n")
	conn <- dbConnect(MySQL(), host="", dbname="", ="", ="")
	result <- dbReadTable(conn, tableName)
	dbDisconnect(conn)
	return(result)
}

getFilesTable <- function() {
	cat("Obtaining files table…\n")
	filesTable <- readData("files")
	filesTable <- subset(filesTable, enabled == 1)
	filesTable$fileId <- as.numeric(as.character(filesTable$fileId))
	
	return (filesTable)
}

getDepsTable <- function() {
	cat("Obtaining deps table…\n")
	depsTable <- readData("deps")
	depsTable <- depsTable[,c("fromFileId", "toFileId", "isolation")]
	depsTable$fromFileId <- as.numeric(as.character(depsTable$fromFileId))
	depsTable$toFileId <- as.numeric(as.character(depsTable$toFileId))
	depsTable$isolation <- as.numeric(as.character(depsTable$isolation))
	
	return (depsTable)
}

getStatsTable <- function() {
	cat("Obtaining stats table…\n")
	statsTable <- readData("stats")
	statsTable <- statsTable[,c("fileId", "vol", "loc", "nm", "nc")]
	statsTable$fileId <- as.numeric(as.character(statsTable$fileId))
	statsTable$vol <- as.numeric(as.character(statsTable$vol))
	statsTable$loc <- as.numeric(as.character(statsTable$loc))
	statsTable$nm <- as.numeric(as.character(statsTable$nm))
	statsTable$nc <- as.numeric(as.character(statsTable$nc))
	
	return(statsTable)	
}

setUpData <- function(filesTable, depsTable) {
	cat("Calculating average incoming isolation…\n")
	incomingIsolation <- aggregate(depsTable$isolation, by=list(depsTable$toFileId), FUN=mean)
	names(incomingIsolation) <- c("toFileId", "incomingIsolation")
	dataset <- merge(filesTable, incomingIsolation, by.x="fileId", by.y="toFileId", all.x=TRUE)
	
	cat("Calculating average outgoing isolation…\n")
	outgoingIsolation <- aggregate(depsTable$isolation, by=list(depsTable$fromFileId), FUN=mean)
	names(outgoingIsolation) <- c("fromFileId", "outgoingIsolation")
	dataset <- merge(dataset, outgoingIsolation, by.x="fileId", by.y="fromFileId", all.x=TRUE)
	
	cat("Calculating indegree…\n")
	indegree <- aggregate(fromFileId ~ toFileId, data=depsTable, FUN="length")
	names(indegree) <- c("toFileId", "indegree")
	dataset <- merge(dataset, indegree, by.x="fileId", by.y="toFileId", all.x=TRUE)
	
	cat("Calculating outdegree…\n")
	outdegree <- aggregate(toFileId ~ fromFileId, data=depsTable, FUN="length")
	names(outdegree) <- c("fromFileId", "outdegree")
	dataset <- merge(dataset, outdegree, by.x="fileId", by.y="fromFileId", all.x=TRUE)
	
	# Isolation rating = number of files that import a certain library. If not the right package prefix, it is 0
	# Therefore, set everything with isolation == 0 to NA
	dataset[!is.na(dataset$incomingIsolation) & dataset$incomingIsolation == 0,]$incomingIsolation <- NA
	dataset[!is.na(dataset$outgoingIsolation) & dataset$outgoingIsolation == 0,]$outgoingIsolation <- NA
	
	# Should be "higher is better" but is reversed in original data
	dataset$incomingIsolation <- 1 - dataset$incomingIsolation
	dataset$outgoingIsolation <- 1 - dataset$outgoingIsolation
	
	return (dataset)
}

getCorrelationMatrix <- function(dataset) {
	result <- data.frame()
	
	BONFERRONI_CORRECTION <- length(numericColumns) * length(numericColumns)
	significanceLevel <- 0.05 / BONFERRONI_CORRECTION
	cat("Using a significance level of", significanceLevel,"\n")
	cat("Creating correlation matrix...\n")

	for (i in 1:ncol(dataset)) {
		currentRow <- list()
		
		for (j in 1:ncol(dataset)) {
			correlation <- cor.test(dataset[,i], dataset[,j], method="spearman", exact=FALSE)
			
			if (j <= i) { 
				if (correlation$p.value >= significanceLevel) star <- "*" else star <- ""
				
				currentCell <- sprintf("%.2f%s", correlation$estimate, star)
			} else {
				if (correlation$p.value == 0)
					currentCell <- "{\tiny \textit{0}}"
				else
					currentCell <- sprintf("\\mbox{{\tiny \textit{%.1e}}}",correlation$p.value)
			}
			
			if (j == ncol(dataset)) sep <- "\\ \\hline" else sep <- "&"
			currentRow <- cbind(currentRow, paste(currentCell, sep))
		}
		
		result <- rbind(result, currentRow)
	}
	
	numericColumns <- getDisplayNames(numericColumns)
	
	names(result) <- numericColumns
	row.names(result) <- numericColumns
	
	return(result)
}

getDisplayNames <- function(numericColumns) {
	numericColumns[1] <- "\textit{PageR.} &"
	numericColumns[2] <- "\textit{Hubb.} &"
	numericColumns[3] <- "\textit{Auth.} &"
	numericColumns[4] <- "\textit{WRM} &"
	numericColumns[5] <- "\textit{CEM} &"
	numericColumns[6] <- "\textit{RCNO} &"
	numericColumns[7] <- "\textit{PNM} &"
	numericColumns[8] <- "\textit{dUn} &"
	numericColumns[9] <- "\textit{dUo} &"
	numericColumns[10] <- "\textit{Maint.} &"
	numericColumns[11] <- "\textit{nU} &"
	numericColumns[12] <- "\textit{nUn} &"
	numericColumns[13] <- "\textit{nUr} &"
	numericColumns[14] <- "\textit{InI.} &"
	numericColumns[15] <- "\textit{OutI.} &"
	numericColumns[16] <- "\textit{InD.} &"
	numericColumns[17] <- "\textit{OutD.} &"
	
	return (numericColumns)	
}

getFlattenedData <- function(filesTable, depsTable) {
	cat("Merging depsTable and filesTable on fromFileId…\n")
	flattened <- merge(depsTable, filesTable, by.x="fromFileId", by.y="fileId", all.X=TRUE)
	flattened <- flattened[,c("fromFileId", "maintainability", "toFileId", 
							  "CEM", "WRM", "PNM", "RCNO", "isolation", "nrRemovedUnits", "nrNewUnits", "nrUnits", "snapshotId")]
	
	names(flattened)[2] <- "fromMaint"
	names(flattened)[4] <- "fromCEM"
	names(flattened)[5] <- "fromWRM"
	names(flattened)[6] <- "fromPNM"
	names(flattened)[7] <- "fromRCNO"
	names(flattened)[8] <- "isolation"
	names(flattened)[9] <- "fromnUr"
	names(flattened)[10] <- "fromnUn"
	names(flattened)[11] <- "fromnU"
	names(flattened)[12] <- "fromSnapshotId"
	
	cat("Merging depsTable and filesTable on toFileId…\n")
	flattened <- merge(flattened, filesTable, by.x="toFileId", by.y="fileId", all.x=TRUE, all.y=FALSE)
	flattened <- flattened[,c("toFileId", "fromFileId", "fromMaint", "maintainability", 
							  "fromCEM", "fromWRM", "fromPNM", "fromRCNO",
							  "fromnUr", "fromnUn", "fromnU", "fromSnapshotId", 
							  "isolation", "CEM", "WRM", "RCNO", "PNM", 
							  "nrRemovedUnits", "nrNewUnits", "nrUnits", "snapshotId")]
	
	names(flattened)[4] <- "toMaint"
	names(flattened)[14] <- "toCEM"
	names(flattened)[15] <- "toWRM"
	names(flattened)[16] <- "toRCNO"
	names(flattened)[17] <- "toPNM"
	names(flattened)[18] <- "tonUr"
	names(flattened)[19] <- "tonUn"
	names(flattened)[20] <- "tonU"
	names(flattened)[21] <- "toSnapshotId"
	
	return(flattened)
}

# This code does not work currently, but idea is clear:
# - calculate the percentile rank for each dependency (0.0 - 1.0)
# - multiply this number with the isolation rating of the dependency
# - sum all these products over all dependencies over a library
# This is the "combined isolation-weighted sum" of all change in dependencies
getAggregateExposure <- function(flattened, isolation) {
	flattened$perctoCEM <- percentilerank(flattened$toCEM)
	flattened$perctoWRM <- percentilerank(flattened$toWRM)
	flattened$perctoPNM <- percentilerank(flattened$toPNM)
	
	flattened$prodIsoltoCEM <- flattened$perctoCEM * ifelse(isolation, flattened$isolation, 1)
	flattened$prodIsoltoWRM <- flattened$perctoWRM * ifelse(isolation, flattened$isolation, 1)
	flattened$prodIsoltoPNM <- flattened$perctoPNM * ifelse(isolation, flattened$isolation, 1)
	
	sumsCEM <- aggregate(flattened$prodIsoltoCEM, by=list(flattened$fromFileId), FUN=sum)
	sumsWRM <- aggregate(flattened$prodIsoltoWRM, by=list(flattened$fromFileId), FUN=sum)
	sumsPNM <- aggregate(flattened$prodIsoltoPNM, by=list(flattened$fromFileId), FUN=sum)
	
	names(sumsCEM) <- c("fromFileId", "sumCEM")
	names(sumsWRM) <- c("fromFileId", "sumWRM")
	names(sumsPNM) <- c("fromFileId", "sumPNM")
	
	filesTable <- merge(filesTable, sumsCEM, by.x="fileId", by.y="fromFileId", all.x=TRUE, all.y=FALSE)
	filesTable <- merge(filesTable, sumsWRM, by.x="fileId", by.y="fromFileId", all.x=TRUE, all.y=FALSE)
	filesTable <- merge(filesTable, sumsPNM, by.x="fileId", by.y="fromFileId", all.x=TRUE, all.y=FALSE)
	
	cat("isolation=",isolation,"\n")
	print(cor.test(filesTable$sumsCEM ~ filesTable$CEM, method="spearman"))
	print(cor.test(filesTable$sumsPNM ~ filesTable$PNM, method="spearman"))
	print(cor.test(filesTable$sumsWRM ~ filesTable$WRM, method="spearman"))
}

percentilerank <- function(x) {
	rx <- rle(sort(x))
	smaller <- cumsum(c(0, rx$lengths))[seq(length(rx$lengths))]
	larger <- rev(cumsum(c(0, rev(rx$lengths))))[-1]
	rxpr <- smaller/(smaller + larger)
	rxpr[match(x, rx$values)]
}

createMaintCategories <- function(flattened) {
	# Divide systems in 4 maintainability categories:
	# 	1) < p25
	#	2) p25 <= x < p50
	#	3) p50 <= x < p75
	#	4) >= p75
	nrGroups <- 4
	percentiles <- quantile(flattened$fromMaint, c(1:(nrGroups-1))/nrGroups, na.rm=TRUE)
	cat("Using the following maintainability percentiles for ANOVA: \n")
	print(percentiles)
	
	flattened[is.na(flattened$fromMaint),"catMaint"] <- 0
	flattened[!is.na(flattened$fromMaint) & flattened$fromMaint < percentiles[1],"catMaint"] <- 1
	flattened[!is.na(flattened$fromMaint) & flattened$fromMaint >= percentiles[1] 
		& flattened$fromMaint < percentiles[2],"catMaint"] <- 2
	flattened[!is.na(flattened$fromMaint) & flattened$fromMaint >= percentiles[2]
		& flattened$fromMaint < percentiles[3],"catMaint"] <- 3
	flattened[!is.na(flattened$fromMaint) & flattened$fromMaint >= percentiles[3],"catMaint"] <- 4
	
	# Required for anova (aov)
	flattened$catMaint <- factor(flattened$catMaint)
	flattened$fromFileId <- factor(flattened$fromFileId)
	
	return (flattened)
}

draw3dplot <- function(flattened) {
	flattened$percfromWRM <- percentilerank(flattened$fromWRM)
	flattened$perctoWRM <- percentilerank(flattened$toWRM)
	scatterplot3d(flattened$percfromWRM, flattened$perctoWRM, flattened$isolation, main="")
}

# =======================
# This script does four different things: 
# 1) Get a flat files table and store it in RData file & csv file
# 2) Get a dependency table and store it in RData file & csv file
# 3) Get sum of products of isolation and stability ratings for each
#    dependency and calculate correlations with and without isolation turned on
# 4) Print a correlation matrix for in paper (not needed any more)
# 5) Print a 3d scatterplot between fromWRM, toWRM and isolation

args <- commandArgs(trailingOnly = TRUE)
stopifnot(length(args) == 1)

filesTable <- getFilesTable()
depsTable <- getDepsTable()
filesTable <- setUpData(filesTable, depsTable)
statsTable <- getStatsTable()

dataset <- filesTable

if (args[1] == "files") {
	write.csv2(filesTable, file="filesTable.csv", row.names=FALSE)	
} else if (args[1] == "deps") {
	flattened <- getFlattenedData(filesTable, depsTable)
	filesTable$lib <- as.numeric(as.factor(paste(filesTable$groupId, filesTable$artifactId)))
	libNumbers <- filesTable[,c("fileId", "lib")]
	flattened <- merge(flattened, libNumbers, by.x="fromFileId", by.y="fileId", all.x=TRUE, all.y=FALSE)
	names(flattened)[22] <- "fromLib"
	flattened <- merge(flattened, libNumbers, by.x="toFileId", by.y="fileId", all.x=TRUE, all.y=FALSE)
	names(flattened)[23] <- "toLib"
	write.csv2(flattened, file="flattened.csv", row.names=FALSE)
} else if (args[1] == "sumsWithIsolation") {
	flattened <- getFlattenedData(filesTable, depsTable)
	getAggregateExposure(flattened, TRUE)	
} else if (args[1] == "sumsWithoutIsolation") {
	flattened <- getFlattenedData(filesTable, depsTable)
	getAggregateExposure(flattened, FALSE)	
} else if (args[1] == "3dplot") {
	flattened <- getFlattenedData(filesTable, depsTable)
	draw3Dplot(flattened)
} else if (args[1] == "matrix") {
	correlationMatrix <- getCorrelationMatrix(dataset)
	print(correlationMatrix)
} else {
	cat("Invalid option provided. Quitting now.\n")
	q("no")
}

save.image() # filesTable used in scatterplots.r
cat("Finished.\n")
